# Created: <2005/07/08>

use XML::DOM;
use Jcode;
use utf8;

$| = 1;

#------------------------------
# 処理
#------------------------------
sub main {
	($outfile =~ /[ \t]/) and die "\$outfile に空白文字がはいっちゃってるので中止。($outfile)\n";

	parse_xml();        # xml を解析し、%sects を構築
	generate_texi();    # texi ファイル生成
	return;             # 以上。
}

#------------------------------
# 入出力ファイル
#------------------------------

# 出力文字セット (Jcode用)
local $out_charset = "sjis";
# 入力文字セット (Jcode用)
local $in_charset = "utf8";

# 入力ファイル名(XML)
local $datafile = "reference.xml";
# 出力ファイル名(texi)
local $outfile = "reference.texi";     # 空白文字($20)を含まないこと。(info の作成に失敗するから)
# 出力ディレクトリ
local $outdir = "./info";

#------------------------------
# 解析結果格納変数
#------------------------------

# セクション分類による全データハッシュ
local %sects = ();
# 目的別索引用
local %tips = ();

#------------------------------

#目次の順番リスト
local @INDEX_LIST = ("データ型",
					 "変数と定数",
					 "制御構造",
					 "パッケージ",
					 "関数",
					 "マクロ",
					 "シンボル",
					 "数値",
					 "文字",
					 "文字列",
					 "シーケンス",
					 "リスト",
					 "ハッシュ",
					 "配列",
					 "チャンク",
					 "評価",
					 "入出力",
					 "ファイルシステム",
					 "エラー",
					 "ウィンドウ",
					 "バッファ",
					 "ミニバッファ",
					 "リージョン",
					 "モード",
					 "シンタックス",
					 "キーマップ",
					 "テキスト",
					 "検索・正規表現",
					 "ダイアログ",
					 "日付・時間",
					 "メニュー",
					 "ファイラ",
					 "ポジション",
					 "プロセス",
					 "システム",
					 "その他",
					 );

# 目次の別名
local %INDEX_NAME_LIST = ("データ型" => "datatypes",
						  "変数と定数" => "variables",
						  "制御構造" => "program",
						  "関数" => "function",
						  "マクロ" => "macro",
						  "シンボル" => "symbols",
						  "パッケージ" => "packages",
						  "数値" => "numbers",
						  "文字" => "characters",
						  "文字列" => "string",
						  "シーケンス" => "sequence",
						  "リスト" => "list",
						  "ハッシュ" => "hash",
						  "配列" => "array",
						  "チャンク" => "chunk",
						  "評価" => "evaluation",
						  "入出力" => "io",
						  "ファイルシステム" => "filesystem",
						  "エラー" => "errors",
						  "ウィンドウ" => "window",
						  "バッファ" => "buffer",
						  "ミニバッファ" => "minibuffer",
						  "リージョン" => "region",
						  "モード" => "mode",
						  "シンタックス" => "syntax",
						  "キーマップ" => "keymap",
						  "メニュー" => "menu",
						  "日付・時間" => "time",
						  "ダイアログ" => "dialog",
						  "システム" => "system",
						  "ファイラ" => "filer",
						  "テキスト" => "text",
						  "プロセス" => "process",
						  "ポジション" => "position",
						  "検索・正規表現" => "search",
						  "その他" => "misc",
						  );

local %TEXI_CMD_REF = ('Function' => '@defun',
					   'Macro' => '@defmac',
					   'Misc' => 'misc: ',
					   'Special Form' => '@defspec',
					   'Tips' => 'tips: ',
					   'Variable' => '@defvar',
					   );

sub parse_xml {
	my $parser = new XML::DOM::Parser;

	# XML Parse
	&report( sprintf( "parsing %s ... ", $datafile), 1);
	my $doc = $parser->parsefile( $datafile);
	&report( "done.\n");

	# データ処理
	&report( "studying data ... ", 1);
	# <book>
	for my $book ( $doc->getChildNodes){
		# <chapter>
		for my $chapter ( $book->getChildNodes){
			next unless ($chapter->getNodeType == ELEMENT_NODE);

			my $title;
			my %chap;
			# 各エレメント
			for my $element ( $chapter->getChildNodes){
				next unless ( $element->getNodeType == ELEMENT_NODE);

				my $name = $element->getNodeName;
				my $data;
				# エレメントの子供 (TEXT_NODEのはず)
				for my $key ( $element->getChildNodes){
					if ( $key->getNodeType == TEXT_NODE){
						$data = encode_text_for_target($key->getData);
						if( $chap{ $name}){
							$chap{ $name} .= "\t" . $data;
						} else {
							$chap{ $name} = $data;
						}
					}
				}
			}#for my $element

			# パースした chapter(1ヶ)を該当セクションにぶち込むのです。
			push( @ { $sects{ $chap{ "section"}}}, \%chap) if %chap;
			# 必要とあらば目的別索引用のハッシュにも。
			push( @ { $tips { $chap{ "section"}}}, \%chap) if ($chap{ "type"} eq "Tips");
		}#for my $chapter
	}
	&report( "done.\n");
}

sub generate_texi {
	move_to_outdir();
	open( OUT, ">$outfile" ) or die $!;
	report( "writing header...", 1);
	make_texi_header();
	report( "done\n");

	report( "writing main menu...", 1);
	make_texi_topmenu();
	report( "done\n");

	report( "writing contents...", 1);
	make_texi_content();
	report( "done\n");

	report( "writing tips menu...", 1);
	make_texi_tips_menu('Function Index', '');       # パラメータ要確認(前後に位置する index ページをしっかり指定する)
	report( "done\n");

	report( "writing closing part...", 1);
	make_texi_closing();
	report( "done\n");

=COMMENT
		report( "checking sections...", 1);
	&check_section( \%INDEX_NAME_LIST, \%sects );
	report( "done\n");
=cut
	}

#==============================================================================#

# 状況出力
sub report{
	my( $str, $filename) = map { Jcode::convert($_, "sjis", $in_charset) } @_;
	printf( STDOUT "%s: ", $0) if $filename;
	printf STDOUT $str;
}

# 変換日
sub timestamp{
	sprintf "Converted from %s by %s at %s\n", $datafile, $0, scalar localtime();
}

# 出力ディレクトリへ移動(なければ作る)
sub move_to_outdir {
	unless( -d $outdir){
		mkdir $outdir, 0755;
		&report( "mkdir $outdir\n", 1);
	}
	chdir $outdir or die $!;
}

sub writeln {
	my $strbuf = sprintf(scalar(shift), @_);
	print OUT Jcode::convert($strbuf, $out_charset, $in_charset), "\n";
}

sub writeln2 {
	my $strbuf = sprintf(scalar(shift), @_);
	$strbuf =~ s/ \*(\w)/ _$1/g;
	$strbuf =~ s/([^@])\*([,:]|$)/$1_$2/g;
	$strbuf =~ s/@xref\{\*([^\}]*)\*\}/@xref\{_$1_\}/;
	$strbuf =~ s/:(\w)/_$1/g;
	$strbuf =~ s/([es][di]):(\w)/$1_$2/g;
	$strbuf =~ s/([es][di]):\*/$1__/g;
	print OUT Jcode::convert($strbuf, $out_charset, $in_charset), "\n";
}

# 一部の文字は加工せにゃならん (`@' と `{' と '}' )
sub encode_text_for_target {
	my $text = shift;
	$text =~ s/(@|{|})/\@$1/g;
	return $text;
}

### [make_texi_header] ##############################
sub make_texi_header {
	my $info_title = "xyzzy Reference";
	my $infofile;
	($info_file = $outfile) =~ s@\.[^./\\]+$@@;
	$info_file .= '.info';
	# header
	writeln '\input texinfo';
	writeln '@c -*- Mode: Texinfo -*-';
	writeln '@c %**start of header';
	writeln '@setfilename %s', $info_file;
	writeln '@settitle %s', $info_title;
	writeln '@documentlanguage ja';
	writeln '@pagesizes 240mm';
	writeln '@paragraphindent none';
	writeln '@defcodeindex mc';
	writeln '@c %**end of header';
	writeln '';
	# file title
	writeln '@titlepage';
	writeln '@title %s', $info_title;
	writeln '@end titlepage';
	writeln '';
	# file summary
	writeln '@ifnottex';
	writeln '@node Top';
	writeln '@top %s', $info_title;
	writeln '@end ifnottex';
	writeln '';
}

### [make_texi_topmenu] ##############################
sub make_texi_topmenu {
	my $secname;
	local $_;
	# find first section written explicitly.
	for $_ (@INDEX_LIST) {
		if (exists( $sects{ $_ } )) {
			$secname = $_;
			last;
		}
	}

	writeln '@menu';

	my $desc_indent = 30;
	my $menufmt = '%1s %-' . $desc_indent . 's%s';
	# sections
	for $secname (@INDEX_LIST) {
		my $doc_exists_mark = exists( $sects{ $secname } ) ? '*' : ' ';
		writeln $menufmt, $doc_exists_mark, get_section_node( $secname ) . '::', $secname;
	}
	# indexes
	writeln '';
	writeln '@ifnotplaintext';
	my @indexes = ('Tips Index::', '目的別索引',
				   'Function Index::', '関数索引',
				   'Variable Index::', '変数索引',
				   'Misc Index::', 'その他索引'
				   );
	my $idx = 0;
	while ($idx <= $#indexes) {
		writeln $menufmt, '*', $indexes[$idx++], $indexes[$idx++];
	}
	writeln '@end ifnotplaintext';
	writeln '@end menu';
	writeln '';
	writeln '@contents';
	writeln '';
}

### [make_texi_content] ##############################
sub make_texi_content {
	report("\n");
	my (@exist_sects, $section);
	for $section (@INDEX_LIST) {
		push @exist_sects, $section if exists( $sects{ $section } );
	}
	for $idx ( 0..$#exist_sects ) {
		$section = $exist_sects[$idx];
		my @chaps = sort sortby_title (@ { $sects{ $section } });
		# section menu
		make_texi_section_menu(
			$section,
			( 0 < $idx ) ? $exist_sects[$idx - 1] : "",
			( $idx < $#exist_sects ) ? $exist_sects[$idx + 1] : "",
			\@chaps);
		# each sections
		for $cidx ( 0..$#chaps ) {
			make_texi_chapter(
				$section,
				$chaps[$cidx],
				( 0 < $cidx ) ? get_chapter_node( $chaps[$cidx - 1] ) : "",
				( $cidx < $#chaps ) ? get_chapter_node( $chaps[$cidx + 1] ) : "",
				);
		}
	}
	report("writing all of sections, ");
}
sub sortby_title {
	( $ { $a } { "title" } ) cmp ( $ { $b } { "title" } );
}
=COMMENT ##############################
sub get_chapter_node {
	my $chap = shift;
	my ($title, $package) = (
		$ { $chap } {"title"},
		$ { $chap } {"package"}
		);
	if ($package) {
		sprintf "%s [%s]", $title, $package;
	} else {
		$title;
	}
}
=cut ##############################
sub get_section_node {
	$ret = $INDEX_NAME_LIST{ scalar(shift) };
	$ret and ($ret .= ' $');
}
sub get_chapter_node {
	my $chap = shift;
	$ { $chap } {"title"};
}
# section menu
sub make_texi_section_menu {
	my ($section, $prev_section, $next_section, $chaps) = @_;
	report("[section] $section\n", 1);
	writeln '@node %s, %s, %s, Top',
	get_section_node( $section ),
	get_section_node( $next_section ),
	get_section_node( $prev_section );
	writeln '@chapter %s', $section;
	writeln '';
	writeln '@menu';
	for my $chap (@ { $chaps } ) {
		writeln '* %s::', get_chapter_node($chap);
	}
	writeln '@end menu';
	writeln '';
}
# item(defun, defmacro, defvar, etc.)
sub make_texi_chapter {
	my ($section, $chapter, $prev_chap_title, $next_chap_title) = @_;
	my ($title, $args, $type, $package, $desc, $seealso, $file) = (
		$ { $chapter } { "title" },
		$ { $chapter } { "arguments" },
		$ { $chapter } { "type" },
		$ { $chapter } { "package" },
		$ { $chapter } { "description" },
		$ { $chapter } { "seealso" },
		$ { $chapter } { "file" }
		);
	my $cmd = $TEXI_CMD_REF{ $type };
	writeln '@node %s, %s, %s, %s', get_chapter_node($chapter), $next_chap_title, $prev_chap_title, get_section_node( $section );
	# heading
	if ($cmd =~ /^@([^ ]+)/) {
		# defun, defmac, defvar and so on.
		my $cmdbody = $1;
		writeln '@unnumberedsec %s', $title;
		if ($cmd =~ /defvar/) {
			writeln '@vindex %s', $title;
			writeln '';
			writeln '%s %s %s', $cmd, $title, $args;
		} else {
			if ($title =~ m/:|^\*$/) {
				writeln '@findex %s', $title;
			}
			writeln '';
			if ($args) {
				if ($title eq substr($args, 0, length($title))) {
					writeln '%s %s', $cmd, $args;
				} else {
					writeln '%s %s %s', $cmd, $title, $args;
				}
			} else {
				writeln '%s %s', $cmd, $title;
			}
		}
	} else {
		# a chapter about tips/misc
		writeln '@unnumberedsec %s%s', @cmd, $title;
		writeln '@mcindex %s', $title;
		writeln '';
	}
	writeln '@format';
	# property
	if ($package or $file) {
		writeln '';
		($package) and (writeln "[Package] %s", $package);
		($file)    and (writeln "[File   ] %s", $file);
	}
	# description
	writeln '%s', $desc;
	writeln '@end format';
	if ($cmd =~ /^@([^ ]+)/) {
		writeln '@end %s', $1;
	}
	# see also
	if ($seealso) {
		writeln '';
		map { writeln '@xref{%s}. @*', $_ } split "\t", $seealso;
	}
	writeln '';
}

### [make_texi_tips_menu] ##############################
sub make_texi_tips_menu {
	my ($next_index, $prev_index) = @_;
	my (@exist_sects, $section);
	for $section (@INDEX_LIST) {
		push @exist_sects, $section if exists( $tips{ $section } );
	}
	writeln '@ifnotplaintext';
	writeln '@node Tips Index, %s, %s, Top', $next_index, $prev_index;
	writeln '@unnumbered Tips Index';
	writeln '';
	writeln '@menu';
	my $desc_indent = 30;
	my $menufmt = '%1s %-' . $desc_indent . 's%s';
	for $idx ( 0..$#exist_sects ) {
		$section = $exist_sects[$idx];
		my @chaps = sort sortby_title (@ { $tips{ $section } });

		# section separater
		writeln '==[%s]==============================', $section;
		# each sections
		for $cidx ( 0..$#chaps ) {
			writeln '* %s::', get_chapter_node( $chaps[$cidx] );
		}
		writeln '';
	}
	writeln '@end menu';
	writeln '';
}

### [make_texi_closing] ##############################
sub make_texi_closing {
	writeln '@node Function Index, Variable Index, Tips Index, Top';
	writeln '@unnumbered Function Index';
	writeln '';
	writeln '@printindex fn';
	writeln '';
	writeln '@node Variable Index, Misc Index, Function Index, Top';
	writeln '@unnumbered Variable Index';
	writeln '';
	writeln '@printindex vr';
	writeln '';
	writeln '@node Misc Index, , Variable Index, Top';
	writeln '@unnumbered Misc Index';
	writeln '';
	writeln '@printindex mc';
	writeln '@end ifnotplaintext';
	writeln '';
	writeln '@bye';
}

main();
0;
